# SPDX-FileCopyrightText: 2026 bartzbeielstein
# SPDX-License-Identifier: AGPL-3.0-or-later

"""Configurator utilities for task-level configuration."""

from spotforecast2.manager.configurator.config_entsoe import ConfigEntsoe

__all__ = ["ConfigEntsoe"]
